#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <limits.h>
#include <errno.h>

#include "config.h"
#include "imapfilter.h"
#include "data.h"


extern int sockpri;
extern account_t *accounts;
extern filter_t *filters;
extern namesp_t nsppri;

unsigned int options;		/* Program options. */
unsigned int flags = 0;		/* Program flags. */
unsigned int capabilities;	/* Capabilities of mail server. */
unsigned int interval = 0;	/* Poll at the specified interval. */
char logfile[PATH_MAX];		/* Log file. */
char *home = NULL;		/* User's home directory. */
uid_t ruid, euid;		/* Real and effective UID. */



/*
 * In the beginning there was main()...
 */
int main(int argc, char *argv[])
{
    int c, r, f = 0;
    pid_t pid;
    char *confile = NULL;	/* Configuration file. */
    account_t *ca;		/* Current account. */
    mbox_t *cm;			/* Current mailbox. */

    ruid = getuid();
    euid = geteuid();
    seteuid(ruid);		/* Drop root privileges. */

    home = getenv("HOME");
    options = (OPTION_DETAILS_NORMAL | OPTION_NAMESPACE | OPTION_WARNING);
    *logfile = 0;

    while ((c = getopt(argc, argv, "c:d:hkl:"
#ifdef ENCRYPTED_PASSWORDS
		       "p"
#endif
		       "qv")) != -1) {
	switch (c) {
	case 'c':
	    confile = optarg;
	    break;
	case 'd':
	    options |= OPTION_DAEMON_MODE;
	    errno = 0;
	    interval = strtoul(optarg, NULL, 10);
	    if (errno)
		interval = 0;
	    break;
	case 'h':
	    usage();
	    exit(ERROR_UNDEFINED);
	    break;
	case 'k':
	    kill_imapfilter();
	    break;
	case 'l':
	    strncat(logfile, optarg, PATH_MAX - 1);
	    break;
#ifdef ENCRYPTED_PASSWORDS
	case 'p':
	    options |= OPTION_PASSWORD_EDITOR;
	    break;
#endif
	case 'q':
	    options &= OPTION_DETAILS_CLEAR;
	    options |= OPTION_DETAILS_QUIET;
	    break;
	case 'v':
	    options &= OPTION_DETAILS_CLEAR;
	    options |= OPTION_DETAILS_VERBOSE;
	    break;
	default:
	    usage();
	    exit(ERROR_UNDEFINED);
	    break;
	}
    }

    create_homedir();

    lockfile_check();
    lockfile_create();

    corefile_disable();

    tty_store();
    catch_signals();

    read_config(confile);

#ifdef ENCRYPTED_PASSWORDS
    read_passwords();

    if ((options & OPTION_PASSWORD_EDITOR)) {
	password_editor();

	secmem_clear();
	lockfile_remove();

	exit(0);
    }
#endif

    open_logfile();

    init_vbuf();


    if (options & OPTION_DAEMON_MODE) {
	f = 1;
	options &= OPTION_DETAILS_CLEAR;
	options |= OPTION_DETAILS_QUIET;
    }
    do {
	for (ca = accounts; ca; ca = ca->next) {

	    if (init_connection(&sockpri, ca->server, ca->port, ca->ssl))
		continue;

	    r = greeting_response(&sockpri);

	    if (r == RESPONSE_BYE || check_capabilities(&sockpri))
		continue;

#ifdef DEBUG
	    test(&sockpri);
#endif

	    if (r != RESPONSE_PREAUTH) {
		if (ca->passwdattr == PASSWORD_NONE) {
		    printf("Enter password for %s@%s: ", ca->username, ca->server);
		    get_password(ca->password, PASSWORD_LEN);
		    ca->passwdattr = PASSWORD_PLAIN;
		}
		if (login(&sockpri, ca->username, ca->password) == RESPONSE_NO) {
		    error("imapfilter: username %s or password rejected at %s\n",
			  ca->username, ca->server);
		    continue;
		}
	    }
	    check_namespace(&sockpri, &nsppri);

	    for (cm = ca->mboxes; cm; cm = cm->next)
		if (!*cm->filters)
		    mailbox_status(&sockpri, cm->name, &nsppri);
		else if (!select_mailbox(&sockpri, cm->name, &nsppri)) {
		    apply_filters(cm->filters);
		    close_mailbox(&sockpri);
		}
	    logout(&sockpri);

	    close_connection(&sockpri);
	}

	/* Fork if in daemon mode. */
	if (f) {
	    f = 0;
	    pid = fork();
	    switch (pid) {
	    case -1:
		fatal(ERROR_FORK, "imapfilter: forking; %s\n", strerror(errno));
		break;
	    case 0:
		secmem_lock();
		setuid(ruid);	/* Capability to regain root privileges will
				   not be needed any more. */
		lockfile_create();
		corefile_disable();
		break;
	    default:
		secmem_clear();
		close_logfile();
		exit(0);
		break;
	    }
	}
	if (interval)
	    sleep(interval);
    } while (options & OPTION_DAEMON_MODE && interval);

    secmem_clear();
    close_logfile();

    lockfile_remove();

    exit(0);
}


/*
 * Print a very brief usage message.
 */
void usage(void)
{
    fprintf(stderr,
	    "usage: imapfilter [-hk"
#ifdef ENCRYPTED_PASSWORDS
	    "p"
#endif
	    "qv] [-c configfile] [-d interval] [-l logfile]\n");
}
